#include "lifegamethread.h"

#include <QDebug>

#include <boost/make_shared.hpp>
#include <boost/foreach.hpp>
#include <boost/random.hpp>
#include <boost/tuple/tuple.hpp>

#include <cmath>
#include <ctime>

#include "stepthread.h"


namespace {

boost::mt19937 random(time(NULL));

boost::uniform_real<> dst01(0.0f, 1.0f);
boost::variate_generator<boost::mt19937, boost::uniform_real<> > rand01(random, dst01);

const float MutationPr = 0.3f;
const float CrossOverPr = 0.5f;

}

namespace LifeGame{

DNA::DNA() :
    genes_(new int[DNAWidth*DNAHeight])
{}

DNA::~DNA(){}

void DNA::mutate(DNA* dst, float pr) const{
    for (int i = 0; i < DNAWidth*DNAHeight; ++i){
        if (rand01() < pr){
            dst->genes_[i] = 1-genes_[i];
        }
        else {
            dst->genes_[i] = genes_[i];
        }
    }
}

void DNA::crossOver(DNA *dst, const DNA &other) const{
    const int x1 = random()%DNAWidth;
    const int x2 = random()%DNAWidth;
    const int y1 = random()%DNAHeight;
    const int y2 = random()%DNAHeight;

    const int leftX   = x1<x2?x1:x2;
    const int rightX  = x1<x2?x2:x1;
    const int topY    = y1<y2?y1:y2;
    const int bottomY = y1<y2?y2:y1;

    for (int x = 0; x < DNAWidth; ++x){
        for (int y = 0; y < DNAHeight; ++y){
            const int i = x + y*DNAWidth;
            if (leftX<=x&&x<rightX && topY<=y&&y<bottomY){
                dst->genes_[i] = other.genes_[i];
            }
            else {
                dst->genes_[i] = genes_[i];
            }
        }
    }
}

void DNA::clone(DNA *dst) const{
    for (int i = 0; i < DNAWidth*DNAHeight; ++i){
        dst->genes_[i] = genes_[i];
    }
}

}

LifegameThread::LifegameThread(QObject *parent) :
    QThread(parent),
    generation_(0),
    swapIndex_(0),
    threads_(2),
    bestAliveNum_(-1.0f),
    bestAliveNum2_(-1.0f),
    stopFlag_(false)
{
    using namespace LifeGame;
    for (unsigned int i = 0; i < threads_.size(); ++i){
        StepThread* p = new StepThread();
        p->setWidth(100);
        p->setHeight(15);
        p->setStepSize(100);
        p->resetField();
        threads_[i] = p;
    }

    for (int i = 0; i < Population; ++i){
        for (int x = 0; x < DNAWidth; ++x){
            for (int y = 0; y < DNAHeight; ++y){
                predators_[i].genes_[x + y*DNAWidth] = random()%2;
                producers_[i].genes_[x + y*DNAWidth] = random()%2;
            }
        }
    }
}

LifegameThread::~LifegameThread(){
}

int LifegameThread::bestAliveNum(){
    int ret;
    mutex_.lock();
    ret = bestAliveNum_;
    mutex_.unlock();
    return ret;
}

LifeGame::DNA LifegameThread::bestDNA(){
    LifeGame::DNA ret;
    mutex_.lock();
    ret = bestDNA_;
    mutex_.unlock();
    return ret;
}

LifeGame::DNA LifegameThread::bestDNA2(){
    LifeGame::DNA ret;
    mutex_.lock();
    ret = bestDNA2_;
    mutex_.unlock();
    return ret;
}

LifeGame::DNA LifegameThread::randomPredator(){
    LifeGame::DNA ret;
    mutex_.lock();
    ret = predators_[(swapIndex_+random()%LifeGame::Population)];
    mutex_.unlock();
    return ret;
}

LifeGame::DNA LifegameThread::randomProducer(){
    LifeGame::DNA ret;
    mutex_.lock();
    ret = producers_[(swapIndex_+random()%LifeGame::Population)];
    mutex_.unlock();
    return ret;
}

void LifegameThread::stop(){
    stopFlag_ = true;
}

void LifegameThread::run(){
    float accumulate = 0.0f;
    float accumulate2 = 0.0f;
    const int dX = threads_.size();
    std::vector<boost::tuples::tuple<int, int> > selectedIndeces(dX);
    std::vector<int> selectedNum(LifeGame::Population);
    std::vector<int> selectedNum2(LifeGame::Population);
    qDebug() << "running..." << swapIndex_;
    for (int count = 0; count < 100; ++count){
        mutex_.lock();
        for (int i = 0; i < LifeGame::Population; ++i){
            results_[i] = 0.0f;
            results_[i] = 0.0f;
            probabilities_[i] = 0.0f;
            probabilities2_[i] = 0.0f;
            selectedNum[i] = 0;
        }
        accumulate = 0.0f;
        accumulate2 = 0.0f;
        // evaluate
        for (int i = 0; i < LifeGame::Population*4; i += dX){
            for (int j = 0; j < dX; ++j){
                const int index1 = random()%LifeGame::Population;
                const int index2 = random()%LifeGame::Population;
                selectedIndeces[j] = boost::tuples::make_tuple(index1, index2);
                ++selectedNum[index1];
                ++selectedNum2[index2];
            }
            int c = 0;
            BOOST_FOREACH(StepThread* it, threads_){
                it->resetField();
                it->setSeed(predators_[selectedIndeces[c].get<0>()].genes_, 1, 0, 0, LifeGame::DNAWidth, LifeGame::DNAHeight);
                it->setSeed(producers_[selectedIndeces[c].get<1>()].genes_, 2, 10, 0, LifeGame::DNAWidth, LifeGame::DNAHeight);
                it->start();
                ++c;
            }
            c = 0;
            BOOST_FOREACH(StepThread* it, threads_){
                it->wait();
                results_[selectedIndeces[c].get<0>()] += it->alivePredatorNum()*(it->eatenNum()/10.0f);
                const int pNum = it->aliveProducerNum();
                if (it->eatenNum()==0){
                    results2_[selectedIndeces[c].get<1>()] += pNum;
                }
                else {
                    results2_[selectedIndeces[c].get<1>()] += pNum;
                }
                //qDebug() << "eaten : " << it->eatenNum();
                ++c;
            }
            //qDebug() << i;
        }
        for (int i = 0; i < LifeGame::Population; ++i){
            if (selectedNum[i] == 0){
                results_[i] = 0;
                results2_[i] = 0;
            }
            else {
                results_[i] /= selectedNum[i];
                results2_[i] /= selectedNum[i];
            }
            const float result = results_[i];
            if (bestAliveNum_ < result){
                bestAliveNum_ = result;
                for (int j = 0; j < LifeGame::DNAWidth*LifeGame::DNAHeight; ++j){
                    bestDNA_.genes_[j] = predators_[swapIndex_+i].genes_[j];
                }
            }
            const float result2 = results2_[i];
            if (bestAliveNum2_ < result2){
                bestAliveNum2_ = result2;
                for (int j = 0; j < LifeGame::DNAWidth*LifeGame::DNAHeight; ++j){
                    bestDNA2_.genes_[j] = producers_[swapIndex_+i].genes_[j];
                }
            }

            accumulate += results_[i];
            accumulate2 += results2_[i];
        }

        // calculate probability
        probabilities_[0] = results_[0]/accumulate;
        probabilities2_[0] = results2_[0]/accumulate2;
        for (int i = 1; i < LifeGame::Population; ++i){
            probabilities_[i] = probabilities_[i-1] + results_[i]/accumulate;
            probabilities2_[i] = probabilities2_[i-1] + results2_[i]/accumulate2;
        }

        generateNextGeneration();
        swapIndex_ = LifeGame::Population-swapIndex_;

        ++generation_;
        sendMessage("generation" + QString::number(generation_)
                  + " average : " + QString::number(accumulate/(LifeGame::Population*4))
                  + ", " + QString::number(accumulate2/(LifeGame::Population*4))
                  + ",, " + QString::number(bestAliveNum_)
                  + ", " + QString::number(bestAliveNum2_));

        mutex_.unlock();

        if (stopFlag_){
            break;
        }
    }
    stopFlag_ = false;
    sendMessage("stopped");
}

void LifegameThread::generateNextGeneration(){
    const int nextSwapIndex = LifeGame::Population-swapIndex_;
    //bestDNA_.clone(&predators_[nextSwapIndex]);
    for (int i = nextSwapIndex; i < LifeGame::Population+nextSwapIndex; ++i){
        const float parentSelector = rand01();
        const float opSelector = rand01();
        LifeGame::DNA* parent = &predators_[LifeGame::Population+swapIndex_-1];
        LifeGame::DNA* child = &predators_[i];

        for (int pi = 0; pi < LifeGame::Population; ++pi){
            if (parentSelector < probabilities_[pi]){
                parent = &predators_[pi+swapIndex_];
                break;
            }
        }

        if (opSelector < MutationPr){
            parent->mutate(child, 0.01f);
        }
        else if (opSelector < MutationPr+CrossOverPr){
            const float parent2Selector = rand01();
            LifeGame::DNA* parent2 = &predators_[LifeGame::Population+swapIndex_-1];

            for (int pi = 0; pi < LifeGame::Population; ++pi){
                if (parent2Selector < probabilities_[pi]){
                    parent2 = &predators_[pi+swapIndex_];
                    break;
                }
            }
            parent->crossOver(child, *parent2);
        }
        else {
            parent->clone(child);
        }
    }
    //bestDNA2_.clone(&producers_[nextSwapIndex]);
    for (int i = nextSwapIndex; i < LifeGame::Population+nextSwapIndex; ++i){
        const float parentSelector = rand01();
        const float opSelector = rand01();
        LifeGame::DNA* parent = &producers_[LifeGame::Population+swapIndex_-1];
        LifeGame::DNA* child = &producers_[i];

        for (int pi = 0; pi < LifeGame::Population; ++pi){
            if (parentSelector < probabilities2_[pi]){
                parent = &producers_[pi+swapIndex_];
                break;
            }
        }

        if (opSelector < MutationPr){
            parent->mutate(child, 0.01f);
        }
        else if (opSelector < MutationPr+CrossOverPr){
            const float parent2Selector = rand01();
            LifeGame::DNA* parent2 = &producers_[LifeGame::Population+swapIndex_-1];

            for (int pi = 0; pi < LifeGame::Population; ++pi){
                if (parent2Selector < probabilities2_[pi]){
                    parent2 = &producers_[pi+swapIndex_];
                    break;
                }
            }
            parent->crossOver(child, *parent2);
        }
        else {
            parent->clone(child);
        }
    }
}
