#include "stepthread.h"

#include <QPainter>
#include <QPixmap>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

namespace {

boost::shared_ptr<QPixmap> image;

}

void initializePixImage(){
    image = boost::make_shared<QPixmap>("pix.png");
}

StepThread::StepThread(QObject *parent) :
    QThread(parent),
    width_(0),
    height_(0),
    oldWidth_(0),
    oldHeight_(0),
    stepSize_(100),
    eat_(0)
{
}

int StepThread::aliveNum() const{
    int ret = 0;
    for (int i = 0; i < width_*height_; ++i){
        if (cells_[i]!=0){
            ++ret;
        }
    }
    return ret;
}

int StepThread::eatenNum() const{
    return eat_;
}

int StepThread::alivePredatorNum() const{
    int ret = 0;
    for (int i = 0; i < width_*height_; ++i){
        if (cells_[i]==1){
            ++ret;
        }
    }
    return ret;
}

int StepThread::aliveProducerNum() const{
    int ret = 0;
    for (int i = 0; i < width_*height_; ++i){
        if (cells_[i]==2){
            ++ret;
        }
    }
    return ret;
}

void StepThread::setStepSize(int stepSize){
    stepSize_ = stepSize;
}

void StepThread::setWidth(int width){
    width_ = width;
}

void StepThread::setHeight(int height){
    height_ = height;
}

void StepThread::resetField(){
    if (oldWidth_!=width_ || oldHeight_!=height_){
        cells_.reset(new int[width_*height_]);
        tmp_.reset(new int[width_*height_]);
        oldWidth_ = width_;
        oldHeight_ = height_;
    }
    for (int i = 0; i < width_*height_; ++i){
        cells_[i] = 0;
        tmp_[i] = 0;
    }
    eat_ = 0;
}

void StepThread::randomSeed(){
    for (int i = 0; i < width_*height_; ++i){
        cells_[i] = qrand()%2;
    }
}

void StepThread::setSeed(const boost::shared_array<int> &seed, int type, int x, int y, int width, int height){
    for (int i = 0; i < width; ++i){
        for (int j = 0; j < height; ++j){
            const int xx = x + i;
            const int yy = y + j;
            if (seed[i + j*width] != 0){
                cells_[xx + yy*width_] = type;
            }
        }
    }
}

void StepThread::draw(QPainter *painter) const{
    const QPen tmp(painter->pen());
    const QBrush tmpb(painter->brush());
    painter->setPen(Qt::red);
    painter->setBrush(Qt::red);
    for (int y = 0; y < height_; ++y){
        for (int x = 0; x < width_; ++x){
            if (cells_[x+y*width_]==1){
                painter->drawPixmap(x*8, y*8, *image, 8, 0, 8, 8);
            }
        }
    }
    painter->setPen(Qt::green);
    painter->setBrush(Qt::green);
    for (int y = 0; y < height_; ++y){
        for (int x = 0; x < width_; ++x){
            if (cells_[x+y*width_]==2){
                painter->drawPixmap(x*8, y*8, *image, 0, 0, 8, 8);
            }
        }
    }
    painter->setPen(Qt::white);
    painter->setBrush(Qt::white);
    for (int y = 0; y < height_; ++y){
        for (int x = 0; x < width_; ++x){
            if (cells_[x+y*width_]==0){
                painter->drawRect(x*8, y*8, 7, 7);
            }
        }
    }
    painter->setPen(tmp);
    painter->setBrush(tmpb);
}

int StepThread::nextState(int x, int y){
    const int* cells = &cells_[0];
    int alive = 0;
    int type1Alive = 0;
    int type2Alive = 0;
    for (int i = x-1; i <= x+1; ++i){
        for (int j = y-1; j <= y+1; ++j){
            if (i!=x || j!=y){
                const int x = i<0?width_-1:i>=width_?0:i;
                const int y = j<0?height_-1:j>=height_?0:j;
                const int cell = cells[x + y*width_];
                if (cell > 0){
                    ++alive;
                    if (cell == 1){
                        ++type1Alive;
                    }
                    else if (cell == 2){
                        ++type2Alive;
                    }
                }
            }
        }
    }
    const int cell = cells[x + y*width_];
    if (cell == 1){
        if (type1Alive == 2 || type1Alive == 3){
            return 1;
        }
        else {
            return 0;
        }
    }
    else if (cell == 2){
        const bool eaten = type1Alive>1;
        if (eaten){
            // Hׂ܂
            ++eat_;
        }
        if (type2Alive == 2 || type2Alive == 3){
            if (eaten){
                return 1;
            }
            else {
                return 2;
            }
        }
        else {
            return 0;
        }
    }
    else {
        if (type1Alive == 3){
            return 1;
        }
        else if (type2Alive == 3){
            return 2;
        }
    }
    return 0;
}

void StepThread::run(){
    for (int i = 0; i < stepSize_; ++i){
        int* tmp = &tmp_[0];
        for (int x = 0; x < width_; ++x){
            for (int y = 0; y < height_; ++y){
                tmp[x + y*width_] = nextState(x, y);
            }
        }
        cells_.swap(tmp_);
    }
}

